""" Schedule -- a Catalog of Sheduled Events """

import Zope
from zLOG import LOG, TRACE, DEBUG, INFO, ERROR

from AccessControl import ClassSecurityInfo
from OFS.ObjectManager import ObjectManager
from OFS.SimpleItem import SimpleItem
from Globals import InitializeClass
from Globals import DTMLFile
from Globals import Persistent
from DateTime import DateTime
from ZPublisher import Client
from ZODB.POSException import ConflictError

from whrandom import random
import threading
import thread
import string


checkEvent = threading.Event()


createSchedulerForm = DTMLFile('dtml/createScheduler', globals())


def createSchedule(self, id='NauScheduler', REQUEST={}):
    """ Create a new Scheldule """

    sch = Schedule(id)
    self._setObject(sch.id, sch)

    if REQUEST is not None:
        return self.manage_main(self, REQUEST, update_menu=1)


def initialize(context):
    context.registerClass(
    Schedule,
      permission = 'Add Schedule',
      constructors = (createSchedulerForm, createSchedule), icon = 'www/Schedule.gif')

    context.registerHelpTitle( 'Scheduler Help' )
    context.registerHelp(directory='help')


class ScheduleData(Persistent):
    def __init__(self):
        self._tasks = {}
        self._logs = {}
        self._p_changed = 1


class Schedule(Persistent, ObjectManager, SimpleItem):
    """ The Schedule (catalog of events). This is also a container. """

    meta_type = 'NauScheduler'

    title='Scheduled Event Catalog'

    manage_options = (
            {'label':'manage', 'action':'manageScheduleForm'},
            {'label':'log', 'action':'logScheduleForm'},
            {'label':'Security', 'action':'manage_access'},
            {'label':'Undo', 'action':'manage_UndoForm'},
            {'label':'Event catalog', 'action':'manageCatalogForm'},
        )+ObjectManager.manage_options

    security = ClassSecurityInfo()

    security.declareProtected('Manage portal', 'administrationScheduleForm')
    administrationScheduleForm = DTMLFile('dtml/administrationSchedule', globals())

    security.declareProtected('Manage portal', 'manageScheduleForm')
    manageScheduleForm = DTMLFile('dtml/manageSchedule', globals())

    security.declareProtected('Manage portal', 'logScheduleForm')
    logScheduleForm = DTMLFile('dtml/logSchedule', globals())

    security.declareProtected('Manage portal', 'propertiesForm')
    propertiesForm = DTMLFile('dtml/propertiesSchedule', globals())

    security.declareProtected('Manage portal', 'manageTaskForm')
    manageTaskForm = DTMLFile('dtml/manageTask', globals())

    security.declareProtected('Manage portal', 'skins_logs')
    skins_logs = DTMLFile('skins/logs', globals())

    security.declareProtected('Manage portal', 'skins_managecSchedule')
    skins_manageSchedule = DTMLFile('skins/manageSchedule', globals())

    security.declareProtected('Manage portal', 'skins_managecTask')
    skins_manageTask = DTMLFile('skins/manageTask', globals())


    def __init__(self, id):
        self.id = id
        self._data = ScheduleData()
        self._deleted = 0
        self._p_changed = 1


    security.declareProtected('Manage portal', 'addTask')
    def addTask(self, method, frequency=None, timeNext=None, timeEnd=None, title='', args=(), kwargs={}, setLog=None):
        """
        returns: task id
        method:method - method, whitch will be call on task event (object will be get from ZODB)
        timeNext: DateTime - next task event time
        frequency: float - task event frequency in seconds
        title: str - task comments
        """
        if frequency is None:
            frequency = 86400
            timeEnd = timeNext + 0.5
            
        if frequency <= 0.0:
            raise 'invalid frequency'
        get_transaction().commit()
        IDs = self._data._tasks.keys()
        while 1:
            taskID = string.zfill( str(int(random()*1000000000)), 9 )
            if taskID not in IDs: break

        nowTime = DateTime()
        if timeNext==None:
            timeNext = nowTime
        if (timeNext < nowTime):
            timeNext = nowTime

        methodName = method.im_func.func_name
        object =  method.im_self

        if setLog:
            thereIsLog = 1
        else:
            thereIsLog = 0

        while 1:
            get_transaction().begin()
            try:
                self._data._tasks[taskID] = {'methodName':methodName, 'timeNext':timeNext, 'timeEnd':timeEnd, 'timeLast':None, 'frequency':frequency, 'status':'resumed',
                                                'title':title, 'physicalPath':'/'.join(object.getPhysicalPath()), 'type':'method',
                                                'args':args, 'kwargs':kwargs, 'thereIsLog':thereIsLog}
                self._data._logs[taskID] = {}
                self._data._p_changed = 1
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)

        checkEvent.set()
        return taskID


    security.declareProtected('Manage portal', 'addPythonScript')
    def addPythonScript(self, pythonScript, frequency, timeNext=None, timeEnd=None, title='', setLog=None):
        if frequency <= 0.0:
            raise 'invalid frequency'
        get_transaction().commit()
        IDs = self._data._tasks.keys()
        while 1:
            taskID = string.zfill( str(int(random()*1000000000)), 9 )
            if taskID not in IDs: break

        nowTime = DateTime()
        if timeNext==None:
            timeNext = nowTime
        if (timeNext < nowTime):
            timeNext = nowTime

        methodName = '__call__'

        while 1:
            get_transaction().begin()
            try:
                self.manage_addProduct['PythonScripts'].manage_addPythonScript(taskID)
                script = self._getOb(taskID)
                script.ZPythonScript_edit('', pythonScript)
                script.manage_proxy(('Manager',))
                object =  script
                self._p_changed = 1
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)


        if setLog:
            thereIsLog = 1
        else:
            thereIsLog = 0

        while 1:
            get_transaction().begin()
            try:
                self._data._tasks[taskID] = {'methodName':methodName, 'timeNext':timeNext, 'timeEnd':timeEnd, 'timeLast':None, 'frequency':frequency, 'status':'resumed',
                                                'title':title, 'physicalPath':'/'.join(object.getPhysicalPath()), 'type':'script', 'thereIsLog':thereIsLog,
                                                 'args':(), 'kwargs':{}}
                self._data._logs[taskID] = {}
                self._data._p_changed = 1
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)

        checkEvent.set()
        return taskID


    security.declareProtected('Manage portal', 'getTaskProperties')
    def getTaskProperties(self, taskID):
        taskProperties = {}
        if not (taskID in self._data._tasks.keys() ):
            return None
        task = self._data._tasks[taskID]

        freq = task['frequency']
        nowTimeSec = DateTime().timeTime()
        timeNextSec = task['timeNext'].timeTime()
        if timeNextSec < nowTimeSec:
            timeNextSec = int( (nowTimeSec - timeNextSec)/freq )*freq + freq + timeNextSec

        taskProperties['timeNext'] = DateTime(float(timeNextSec))
        taskProperties['timeEnd'] = task.get('timeEnd', None)
        taskProperties['frequency'] = task['frequency']
        taskProperties['title'] = task['title']
        taskProperties['physicalPath'] = task['physicalPath']
        taskProperties['status'] = task['status']
        taskProperties['type'] = task['type']
        taskProperties['thereIsLog'] = task['thereIsLog']
        return taskProperties


    security.declareProtected('Manage portal', 'setTaskProperties')
    def setTaskProperties(self, taskID, frequency=None, timeNext=None, title=None, physicalPath=None, args=None, kwargs=None):
        get_transaction().commit()
        task = self._data._tasks[taskID]
        while 1:
            get_transaction().begin()
            try:
                if frequency is not None:
                    if frequency <= 0.0:
                        raise 'invalid frequency'
                    task['frequency'] = frequency
                    self.ignoreLater((taskID,))
                if timeNext is not None:
                    task['timeNext'] = timeNext
                    self.ignoreLater((taskID,))
                if title is not None:
                    task['title'] = title
                if args is not None:
                    task['args'] = args
                if kwargs is not None:
                    task['kwargs'] = kwargs
                if physicalPath is not None:
                    task['physicalPath'] = physicalPath
                self._data._p_changed = 1
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)

        checkEvent.set()


    security.declareProtected('Manage portal', 'delTask')
    def delTask(self, taskID):
        get_transaction().commit()
        while 1:
            get_transaction().begin()
            try:
                if self._data._tasks[taskID]['type'] == 'script':
                    self._delObject(taskID)
                self._p_changed = 1
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)

        while 1:
            get_transaction().begin()
            try:
                del self._data._tasks[taskID]
                del self._data._logs[taskID]
                self._data._p_changed = 1
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)
        checkEvent.set()


    security.declareProtected('Manage portal', 'checkDaemon')
    def checkDaemon(self):
        """
        returns: 1 if Daemon exists else 0
        """
        exist = 0
        path = self._path
        for thr in threading.enumerate():
            if thr.getName()=='NauScheduler_'+path:
                exist = 1
        return exist


    security.declareProtected('Manage portal', 'startDaemon')
    def startDaemon( self, postpone=None ):
        """
        start Daemon for instanse Schedule if it has not started yet
        returns: 1 if Daemon exists,
                 0 if Daemon started successful

        """
        path = self._path
        exist = 0
        for thr in threading.enumerate():
            if thr.getName()=='NauScheduler_'+path:
                exist = 1
        if exist:
            return 1
        dispatcher = threading.Thread(target=dispatcherProc, name='NauScheduler_'+path, args=(path,))
        dispatcher.setDaemon(1)
	if postpone:
	    return dispatcher
	dispatcher.start()
        del dispatcher
        return 0


    security.declareProtected('Manage portal', 'ignoreLater')
    def ignoreLater(self, taskIDs=[]):
        nowTime = DateTime()
        if taskIDs==[]:
            taskIDs = self._data._tasks.keys()
        while 1:
            get_transaction().begin()
            try:
                for taskID in taskIDs:
                    task = self._data._tasks[taskID]
                    if task['status'] == 'paused':
                        continue
                    if (task['timeNext'] < nowTime):
                        task['timeNext'] = nowTime
                self._data._p_changed = 1
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)


    security.declareProtected('Manage portal', 'taskPause')
    def taskPause(self, taskID):
        get_transaction().commit()
        while 1:
            get_transaction().begin()
            try:
                self._data._tasks[taskID]['status'] = 'paused'
                self._data._p_changed = 1
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)
        checkEvent.set()


    security.declareProtected('Manage portal', 'taskResume')
    def taskResume(self, taskID):
        get_transaction().commit()
        while 1:
            get_transaction().begin()
            try:
                self._data._tasks[taskID]['status'] = 'resumed'
                self._data._p_changed = 1
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)
        self.ignoreLater((taskID,))
        checkEvent.set()


    security.declareProtected('Manage portal', 'manage_tasks')
    def manage_tasks(self, year=None, month=None, day=None, hour=None, minute=None, second=None, 
                           end_year=None, end_month=None, end_day=None, end_hour=None, end_minute=None, end_second=None,
                           freq_day=None, freq_hour=None, freq_minute=None, freq_second=None,
                        pythonScript='', title='', toDel=[], taskID=None, setLog=None, REQUEST=None):
        """ """
        message = ''

        if REQUEST.has_key('add'):
            self.addPythonScript(pythonScript,
                    float(freq_day)*86400.0+float(freq_hour)*3600.0+float(freq_minute)*60+float(freq_second),
                    DateTime(int(year), int(month), int(day), int(hour), int(minute), int(second)),
                    DateTime(int(end_year), int(end_month), int(end_day), int(end_hour), int(end_minute), int(end_second)),
                    title, setLog)
        elif REQUEST.has_key('modif'):
            self.taskAdd(title, pythonScript, timeNext, (int(timeFrequency[0]), int(timeFrequency[1]), int(timeFrequency[2])), description, taskID)
        elif REQUEST.has_key('del'):
            for taskID in toDel:
                self.delTask(taskID)
        elif REQUEST.has_key('start'):
            for taskID in toDel:
                self.taskResume(taskID)
        elif REQUEST.has_key('stop'):
            for taskID in toDel:
                self.taskPause(taskID)
        elif REQUEST.has_key('startDaemon'):
            exist = self.startDaemon()
            if exist:
                message = 'Daemon alredy exists'
            else:
                message = 'Daemon started successfull'


        if REQUEST is not None:
            REQUEST.RESPONSE.redirect( REQUEST['HTTP_REFERER'] )
#       if REQUEST:
#           return self.manageScheduleForm(self, REQUEST, manage_tabs_message=message)


    security.declareProtected('Manage portal', 'getLog')
    def getLog(self, taskID):
        """ """
        get_transaction().commit()
        while 1:
            get_transaction().begin()
            try:
                resultLog = []
                for i in self._data._logs[taskID].keys():
                    resultLog.append(self._data._logs[taskID][i])
                resultLog.sort()
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)

        return resultLog

    security.declareProtected('Manage Portal', 'getTasksIDs')
    def getTasksIDs(self):
        get_transaction().commit()
        while 1:
            get_transaction().begin()
            try:
                IDs = self._data._tasks.keys()
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)
        return IDs


    security.declareProtected('Manage portal', 'manage_afterAdd')
    def manage_afterAdd(self, item, container):
        self._path = '/'.join(item.getPhysicalPath())
        get_transaction().commit()
        self.startDaemon()
        loop_app = self.getPhysicalRoot()
        if not hasattr(loop_app, 'ScheduleList'):
            try: loop_app.manage_addProduct['NauScheduler'].createScheduleList()
	    except: return
            get_transaction().commit()

        scheduleList = getattr(loop_app, 'ScheduleList')
        while 1:
            get_transaction().begin()
            try:
                scheduleList.registrySchedule( self._path )
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)


    security.declareProtected('Manage portal', 'manage_beforeDelete')
    def manage_beforeDelete(self, item, container):
        loop_app = self.getPhysicalRoot()
        if hasattr(loop_app, 'ScheduleList'):
            scheduleList = getattr(loop_app, 'ScheduleList')
            while 1:
                get_transaction().begin()
                try:
                    self._p_changed = 1
                    self._deleted = 1
                    scheduleList.unregistrySchedule( self._path )
                    get_transaction().commit()
                    break
                except ConflictError:
                    threading.Event().wait(random()*1.0)
        checkEvent.set()
        

def taskProc(schedulePath, taskRecord, taskID):
    import sys, StringIO, traceback
    thereIsLog = taskRecord['thereIsLog']

    if thereIsLog:
        log_app = Zope.app()
        try:
            schedule = log_app.unrestrictedTraverse(schedulePath)
            IDs = schedule._data._logs[taskID].keys()
        except:
            log_app._p_jar.close()
            del log_app
            return

        while 1:
            runID = string.zfill( str(int(random()*1000000000)), 9 )
            if runID not in IDs: break
        while 1:
            get_transaction().begin()
            try:
                schedule._data._logs[taskID][runID] = {'begin': DateTime()}
                schedule._data._p_changed = 1
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)
            except:
                log_app._p_jar.close()
                del log_app
                return
        log_app._p_jar.close()
        del log_app

    app = Zope.app()
    taskPath = taskRecord['physicalPath']
    task = app.unrestrictedTraverse(taskPath)

    method = getattr(task, taskRecord['methodName'])
    args = taskRecord['args']
    kwargs = taskRecord['kwargs']



    error = None
    response = ''

    try:

        import AccessControl.SecurityManagement, AccessControl.SpecialUsers, AccessControl.User
        system=AccessControl.User.UnrestrictedUser('System Processes','',('manage', 'Member','Manager',), [])
        AccessControl.SecurityManagement.newSecurityManager(None, system)

        response = apply(method, args, kwargs)
        del system
        get_transaction().commit()
    except:
        errorIO = StringIO.StringIO()
        traceback.print_exc(10, errorIO)
        error = errorIO.getvalue()

    app._p_jar.close()
    del app

    log_app = Zope.app()
    try:
        schedule = log_app.unrestrictedTraverse(schedulePath)
        IDs = schedule._data._logs[taskID].keys()
    except:
        log_app._p_jar.close()
        del log_app
        return
    while 1:
        get_transaction().begin()
        try:
            schedule._data._tasks[taskID]['once_running'] = 1
            schedule._data._p_changed = 1
            get_transaction().commit()
            break
        except ConflictError:
            threading.Event().wait(random()*1.0)
        except:
            log_app._p_jar.close()
            del log_app
            return
 
    log_app._p_jar.close()
    del log_app

    if thereIsLog:
        log_app = Zope.app()
        try:
            schedule = log_app.unrestrictedTraverse(schedulePath)
            IDs = schedule._data._logs[taskID].keys()
        except:
            log_app._p_jar.close()
            del log_app
            return
        while 1:
            get_transaction().begin()
            try:
                schedule._data._logs[taskID][runID]['end'] = DateTime()
                schedule._data._logs[taskID][runID]['response'] = response
                schedule._data._logs[taskID][runID]['error'] = error
                schedule._data._p_changed = 1
                get_transaction().commit()
                break
            except ConflictError:
                threading.Event().wait(random()*1.0)
            except:
                log_app._p_jar.close()
                del log_app
                return
        log_app._p_jar.close()
        del log_app


def dispatcherProc(schedulePath):
    app = Zope.app()
    try:
        schedule = app.unrestrictedTraverse(schedulePath)
    except:
        app._p_jar.close()
        del app
        return
    tasks = tasksCopy(schedule._data._tasks)

    while 1:
        interval=60
        taskID, typ = tasks.getEarliestTask()
        if taskID is not None:
            if typ == 'onRun':
                timeNext = tasks[taskID]['timeNext']
            else:
                timeNext = tasks[taskID]['timeEnd']
            
        if taskID is None:
            pass
        elif timeNext.isFuture():
            ainterval= timeNext.timeTime() - DateTime().timeTime()
            if ainterval < float(0): ainterval=float(0)
            interval=ainterval
        else:
            if typ == 'onDel':
                if tasks.has_key(taskID):
                    del tasks[taskID]
                schedule.delTask(taskID)
            else:
                tasks[taskID]['timeNext'] = DateTime( float(tasks[taskID]['timeNext'].timeTime() + tasks[taskID]['frequency']) )
    
                task = threading.Thread(target=taskProc, args=(schedulePath, tasks[taskID], taskID),
                                     name='task_%s_%s'%(schedulePath, taskID))
                task.setDaemon(1)
                task.start()
                del task

            interval=float(0)


        checkEvent.wait(interval)
        if checkEvent.isSet():
            threading.Event().wait(5)
            checkEvent.clear()
            app._p_jar.close()
            del app
            app = Zope.app()
            try:
                schedule = app.unrestrictedTraverse(schedulePath)
            except:
                app._p_jar.close()
                del app
                return
            tasks.reload(schedule._data._tasks)

    app._p_jar.close()
    del app
    return


class tasksCopy:
    def __init__(self, tasks):
        self.reload(tasks)


    def __getitem__(self, item):
        return self._tasks[item]

    def has_key(self, key):
        return self._tasks.has_key(key)


    def __delitem__(self, key):
        del self._tasks[key]


    def getEarliestTask(self):
        timeNext=None
        idNext = None
        typ = None
        for taskID in self._tasks.keys():
            if self._tasks[taskID]['status'] == 'paused':
                continue
            freq = self._tasks[taskID]['frequency']
            if timeNext==None  or  timeNext > self._tasks[taskID]['timeNext']:
                timeNext=self._tasks[taskID]['timeNext']
                idNext=taskID
                typ = 'onRun'

            if timeNext==None  or  (self._tasks[taskID].get('timeEnd', None)!=None and timeNext > self._tasks[taskID]['timeEnd']):
                timeNext=self._tasks[taskID]['timeEnd']
                idNext=taskID
                typ = 'onDel'
        return (idNext, typ)


    def reload(self, tasks):
        import copy
        self._tasks = copy.deepcopy(tasks)

        nowTimeSec = DateTime().timeTime()
        for taskID in self._tasks.keys():
            freq = self._tasks[taskID]['frequency']
            timeNextSec = self._tasks[taskID]['timeNext'].timeTime()
            if timeNextSec < nowTimeSec and self._tasks[taskID].get('once_running', 0):
                timeNextSec = int( (nowTimeSec - timeNextSec)/freq )*freq + freq + timeNextSec
                self._tasks[taskID]['timeNext'] = DateTime(float(timeNextSec))


InitializeClass(Schedule)
